/* zklBigNum.c : mega digit ints
 * Base on MPIR: http://www.mpir.org/
 *   or GMP (GNU Multiple Precision Arithmetic Library): http://gmplib.org/
 * 
 * Manual:
 *    https://gmplib.org/manual/
 *    https://gmplib.org/gmp-man-6.1.2.pdf
 * 
 * Compile multi-threaded (MSVC properties/C/Code Generation/Runtime library)
 * Linked against zkl.lib (zkl exports)
 *    msvcrt.lib.
 *    In MSVC properties/Linker/Input, set "Ignore All Default Libraries" to
 *    "Yes (/NODEFAULTLIB)". This creates a much smaller DLL and is why you
 *    need to link against msvcrt.lib
 * 
 * Thread safe I think
 * 
 * Copyright (c) 2010 Craig Durland craigd@zenkinetic.com
 * License: The zlib/libpng License,
 * 	http://www.opensource.org/licenses/zlib-license.php
 * 
 * The MPIR library is covered by LGPL3, which may supercede my license.
 * Ask somebody who knows if it matters.
 */

#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <time.h>	// clock()
#ifdef __unix__
   #include <gmp.h>	// mpir.h or gmp.h
#else
//   #include "../../../mpir-2.1.1/mpir.h"	// Windows
   #include "mpir.h"	// Windows
#endif

#include "zklObject.h"	
#include "zklImports.h"
#include "zklAtomic.h"
#include "zklData.h"
#include "zklList.h"
#include "zklMemory.h"
#include "zklString.h"
#include "zklUtil.h"

static Instance *BN_create(Instance *,pArglist,pVM);

static int bigIntId;

#if 0
static int bigFloatId;

typedef struct 
{
   Instance instance;
   mpf_t    bf;
} BigFloat;

static ZKL_Object BigFloatObject;
#endif

/* ******************************************************************** */
/* **************************** Big Ints ****************************** */
/* ******************************************************************** */

typedef struct 
{
   BInstance instance;
   mpz_t     bi;
} BigInt;

#define BIbi(i) ( ((BigInt *)i)->bi )

static ZKL_Object BigIntObject;

static IBucketHeader *biBuckets;

static int biFree(Instance *self) { mpz_clear(((BigInt *)self)->bi); return 1; }


    /* BigInt.toString() --> convert to base 10 integer
     * BigInt.toString(base) where base is 2..62, -2..-36
     * BigInt.toString(base,n,c) n is commas-every-n, c is comma char
     * For base in the range 2..36, digits and lower-case letters are used;
     *   for -2..-36, digits and upper-case letters are used; for 37..62,
     *   digits, upper-case letters, and lower-case letters (in that
     *   significance order) are used.
     */
static Instance *BI_toString(BigInt *self,pArglist arglist,pVM vm)
{
   char	     c, *ptr;
   int	     n, cat, base = 10;
   int64_t   i64;
   Instance *r;

   if (arglistTryToGetInt(arglist,0,&i64,0,vm))
   {
      base = (int)i64;
      if (!( (base>=2 && base<=62) || (base>=-36 && base<=-2) ))
      {
	 char buf[100];
	 sprintf(buf,"base(%d): range is 2 .. 62, -36..-2",base);
	 vmThrow(vm,E_VALUE_ERROR,buf);
     }
   }

   cat = 0;
   if (arglistTryToGetInt(arglist,1,&i64,0,vm))	// commas
   {
      cat = (int)i64;
      c   = *arglistGetOnlyString(arglist,2,0,vm);
      if (cat < 0) cat = 0;
   }

   n = mpz_sizeinbase(self->bi, abs(base)) + 2;  // "-" & "\0"
   if (cat) n += n/cat;
   r = stringAllocate(n,&ptr,I_OWNED,vm);
   mpz_get_str(ptr,base,self->bi);
   if (cat) fmtCommaize(ptr,cat,c);
   return r;
}

    // .toIntString(base [,commasAt,c]) --> .toString(base [,commasAt,c])
static Instance *BI_toIntString(BigInt *self,pArglist arglist,pVM vm)
   { return BI_toString(self,arglist,vm); }

    // .toData([endian-ness]) --> bytes
static Instance *BI_toData(BigInt *self,pArglist arglist,pVM vm)
{
   size_t    sz;
   int       endian = 0, order = 1;  // order is endian-ness of the blob
   Byte     *bytes;
   Instance *d;
   
   sz    = (mpz_sizeinbase(self->bi,2) + 8-1)/8;
   d     = dataCreate(sz,I_OWNED,vm);
   bytes = dataWillFill(d);
   mpz_export(bytes,&sz,order,1,endian,0,self->bi);
   dataSetSize(d,sz);

   return d;
}

/* 
   size_t mpz_sizeinbase (mpz_t op, int base)
   Return the size of op measured in number of digits in the given base.
   base can vary from 2 to 62.  The sign of op is ignored, just the
   absolute value is used.  The result will be either exact or 1 too
   big.  If base is a power of 2, the result is always exact.  If op is
   zero the return value is always 1.
   >>>>>>>>>>>> note the "1 too big"
 */
static int biNumDigits(mpz_t sn,int base,pVM vm)
{
   int   cnt = 0;
   mpz_t n;

   if(!mpz_cmp_ui(sn,0)) return 1;	// "0" has 1 digit

   mpz_init_set(n,sn);
   do{
      cnt++;
      mpz_tdiv_q_ui(n,n,10);
   }while(mpz_cmp_ui(n,0));
   mpz_clear(n);	// GC
   return cnt;
}


    // BI.len([base=10]) --> # digits
static Instance *BI_len(BigInt *self,pArglist arglist,pVM vm)
{
#if 0
   int	   n, base = 10;
   int64_t i64;

   if (arglistTryToGetInt(arglist,0,&i64,0,vm))
   {
      base = (int)i64;
      if (base < 2 || base > 36) base = 10;
   }

   n = mpz_sizeinbase(self->bi, base);
   return intCreate(n,vm);
#else
   return intCreate(biNumDigits(((BigInt *)self)->bi,10,vm),vm);
#endif
}

static __inline int64_t zToI64(mpz_t zz)
{
   #if (LONG_MAX == 2147483647L)	// 32 bit ints
      int     m = 0;
      int64_t n;
      mpz_t   z;

      if (mpz_cmpabs_ui(zz,ULONG_MAX) < 0)	// fits in 32 bits
	 return mpz_get_si(zz);

      mpz_init_set(z,zz);
      if (mpz_cmp_ui(z,0) < 0) { mpz_neg(z,z); m = 1; }
      n = mpz_get_ui(z);
      mpz_tdiv_q_2exp(z,z,32);	// z >> 32
      n |= (int64_t)mpz_get_ui(z) << 32;
      if(m) n = -n;
      mpz_clear(z);
      return n;
   #else				// 64 bit ints
      return mpz_get_si(zz);
   #endif
}

    // .toInt()
static Instance *BI_toInt(BigInt *self,pArglist arglist,pVM vm)
   { return intCreate(zToI64(self->bi),vm); }

    // .toFloat()
static Instance *BI_toFloat(BigInt *self,pArglist arglist,pVM vm)
   { return floatCreate(mpz_get_d(self->bi),vm); }

    // .unbox(), try to down convert to a 64 bit int
static Instance *BI_unbox(BigInt *self,pArglist arglist,pVM vm)
{
   if (mpz_fits_slong_p(self->bi)) return intCreate(mpz_get_si(self->bi),vm);
   return (Instance *)self;
}

#if 0
    // .toBFloat()
static Instance *BI_toBFloat(BigInt *self,pArglist arglist,pVM vm)
{
   BigFloat *f = (BigFloat *)
		   instanceAllocate(sizeof(BigFloat),&BigFloatObject,1,vm);
   mpf_init(f->bf); mpf_set_z(f->bf,self->bi);
   return addToCollectables((Instance *)f,I_OWNED,vm);
}
#endif

    // .toBool
static Instance *BI_toBool(BigInt *self,pArglist arglist,pVM vm)
   { return mpz_cmp_ui(self->bi,0) ?  BoolTrue : BoolFalse; }

    // .abs
static Instance *BI_abs(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);

   mpz_init(r->bi);
   mpz_abs(r->bi,self->bi);
//   return addToCollectables((Instance *)r,I_OWNED,vm);
   return (Instance *)r;
}

static __inline BigInt *   //-->allocs new [blank] BI, decodes X
eyeballIX(Instance *X, int64_t *N, BigInt **pN, int alloc, pVM vm)
{
   BigInt *bi;

   *pN = 0;
   if (IS_PtrInt(X))		   *N  = PtrInt_TO_N(X);
   else if (TYPEO1(X) == IntType)  *N  = INT64V(X);
   else if (X->objID  == bigIntId) *pN = (BigInt *)X;
   else if (TYPEO1(X) == BoolType) *N  = BOOLV(X);
   else vmThrow(vm,E_TYPE_ERROR,"BigInt.op ?");

   if (!alloc) return 0;

   bi = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_init(bi->bi);
   
   return bi;
}

static __inline BigInt *
getBI(pArglist arglist, int64_t *N, BigInt **pN, pVM vm)
   { return eyeballIX(arglistGet(arglist,0,0,vm),N,pN,1,vm); }

    // NOTE!! you have to mpz_clear(z)!
static __inline void i64ToZ(int64_t i64, mpz_t z)
{
   #if (LONG_MAX == 2147483647L)	// 32 bit ints
      int      m = 0;
      uint32_t h;
      mpz_t    x;

      if (i64 < 0) { i64 = -i64; m = 1; }
      mpz_init_set_ui(z,i64 & 0xffFFffFF);
      if ((h = (i64 >> 32)))
      {
	 mpz_init_set_ui(x,h);
	 mpz_mul_2exp(x,x,32); mpz_add(z,z,x);
      }
      if (m) mpz_neg(z,z);
   #else
      mpz_init_set_si(z,i64);
   #endif
}

   // initialize and set z, you need to clear z
static void x2mpz(Instance *X, mpz_t z, pVM vm)
{
   if (IS_PtrInt(X))		  i64ToZ(PtrInt_TO_N(X),z);
   else if (TYPEO1(X) == IntType) i64ToZ(INT64V(X),z);
   else if (X->objID == bigIntId) mpz_init_set(z,((BigInt *)X)->bi);
   else vmThrow(vm,E_TYPE_ERROR,"BigInt.op ?");
}

    // bi.bitAnd(n) --> bi(self & n)
static Instance *BI_bitAnd(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pM;
   int64_t M;

   r = getBI(arglist,&M,&pM,vm);
   if (pM) mpz_and(r->bi, self->bi,pM->bi);
   else
   {
      mpz_t mask;
      i64ToZ(M,mask);
      mpz_and(r->bi, self->bi,mask); mpz_clear(mask);
   }
   return (Instance *)r;
}

    // bi.bitOr(n) --> bi(self | n)
static Instance *BI_bitOr(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pM;
   int64_t M;

   r = getBI(arglist,&M,&pM,vm);
   if (pM) mpz_ior(r->bi, self->bi,pM->bi);
   else
   {
      mpz_t mask;
      i64ToZ(M,mask);
      mpz_ior(r->bi, self->bi,mask); mpz_clear(mask);
   }
   return (Instance *)r;
}

    // bi.bitXor(n) --> bi(self xor n)
static Instance *BI_bitXor(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pM;
   int64_t M;

   r = getBI(arglist,&M,&pM,vm);
   if (pM) mpz_xor(r->bi, self->bi,pM->bi);
   else
   {
      mpz_t mask;
      i64ToZ(M,mask);
      mpz_xor(r->bi, self->bi,mask); mpz_clear(mask);
   }
   return (Instance *)r;
}

    // bi.shiftLeft(n) --> bi << n
static Instance *BI_shiftLeft(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pn;
   int64_t n;

   r = getBI(arglist,&n,&pn,vm);
   if (pn) mpz_mul_2exp(r->bi, self->bi,mpz_get_ui(pn->bi));
   else    mpz_mul_2exp(r->bi, self->bi,(unsigned long)n);

   return (Instance *)r;
}

    // bi.shiftRight(n) --> bi >> n
static Instance *BI_shiftRight(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pn;
   int64_t n;

   r = getBI(arglist,&n,&pn,vm);
   if (pn) mpz_tdiv_q_2exp(r->bi, self->bi,mpz_get_ui(pn->bi));
   else    mpz_tdiv_q_2exp(r->bi, self->bi,(unsigned long)n);

   return (Instance *)r;
}

    // bi.bitNot() --> bi(~self)
static Instance *BI_bitNot(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_init(r->bi);
   mpz_com(r->bi,self->bi);
   return (Instance *)r;
}

    // bi.gcd(n) --> bi(self gcd N)
static Instance *BI_gcd(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pN;
   int64_t N;

   r = getBI(arglist,&N,&pN,vm);
   if (pN) mpz_gcd(r->bi, self->bi,pN->bi);
   else
   {
      mpz_t z;
      i64ToZ(N,z);
      mpz_gcd(r->bi, self->bi,z); mpz_clear(z);
   }
   return (Instance *)r;
}

    // bi.gcdExt(n) --> T(gcd(self,n),s,t)
    // gcd(a,b) == a*s + b*t
static Instance *BI_gcdExt(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pN, *s,*t;
   int64_t N;

//!!!! need to fence
   s = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   t = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_init(s->bi); mpz_init(t->bi);

   r = getBI(arglist,&N,&pN,vm);
   if (pN) mpz_gcdext(r->bi,s->bi,t->bi, self->bi,pN->bi);
   else
   {
      mpz_t z;
      i64ToZ(N,z);
      mpz_gcdext(r->bi,s->bi,t->bi, self->bi,z); mpz_clear(z);
   }
   return tupleCreateX(vm,(Instance *)r,(Instance *)s,(Instance *)t,ZNIL);
}

    // bi.lcm(n) --> bi(self lcm N)
static Instance *BI_lcm(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pN;
   int64_t N;

   r = getBI(arglist,&N,&pN,vm);
   if (pN) mpz_lcm(r->bi, self->bi,pN->bi);
   else
   {
      mpz_t z;
      i64ToZ(N,z);
      mpz_lcm(r->bi, self->bi,z); mpz_clear(z);
   }
   return (Instance *)r;
}

    // bi.pow(n) --> bi(self ^ n)
static Instance *BI_pow(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pN;
   int64_t N;

   r = getBI(arglist,&N,&pN,vm);
   if (pN) mpz_pow_ui(r->bi, self->bi,mpz_get_ui(pN->bi));
   else    mpz_pow_ui(r->bi, self->bi,(unsigned long)N);
   return (Instance *)r;
}

    // bi.powm(n,m) --> bi((self ^ n) % m)
static Instance *BI_powm(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_t   n,m;

   mpz_init(r->bi);
   x2mpz(arglistGet(arglist,0,0,vm),n,vm);
   x2mpz(arglistGet(arglist,1,0,vm),m,vm);
   mpz_powm(r->bi, self->bi,n,m);
   mpz_clears(n,m,0);
//   else    mpz_powm_ui(r->bi, self->bi,(unsigned long)N);
   return (Instance *)r;
}

    // bi.root(n) --> bi(nthRoot(self))
static Instance *BI_root(BigInt *self,pArglist arglist,pVM vm)
{
   BigInt *r, *pN;
   int64_t N;

   r = getBI(arglist,&N,&pN,vm);
   if (pN) mpz_root(r->bi, self->bi,mpz_get_ui(pN->bi));
   else    mpz_root(r->bi, self->bi,(unsigned long)N);
   return (Instance *)r;
}

    // bi.sqrt(n) --> bi
static Instance *BI_sqrt(BigInt *self,pArglist arglist,pVM vm)
{
//!!!??? use mpz_sqrt???
   MLIST(mlist,1);
   mlistBuild(mlist,intCreate(2,vm),ZNIL);
   return BI_root(self,(Instance *)mlist,vm);
}

    // .div2(d) --> T(q,r)
static Instance *BI_div2(BigInt *self,pArglist arglist,pVM vm)
{
   int64_t  N;
   BigInt  *pN, *r, *q = getBI(arglist,&N,&pN,vm);

   if (pN) // d is a BigInt
   {
      if (0 == mpz_cmp_ui(pN->bi,0)) vmThrow(vm,E_MATH_ERROR,"Divide by zero");
      r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
      mpz_init(r->bi);
      mpz_tdiv_qr(q->bi,r->bi,self->bi,pN->bi);
   }
   else
   {
      mpz_t d;
      if (N == 0) vmThrow(vm,E_MATH_ERROR,"Divide by zero");
      i64ToZ(N,d);
      r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
      mpz_init(r->bi);
      mpz_tdiv_qr(q->bi,r->bi, self->bi,d); mpz_clear(d);
   }
   return tupleCreateX(vm,(Instance *)q,(Instance *)r,ZNIL);
}

    // bi.factorial() --> bi!
static Instance *BI_factorial(BigInt *self,pArglist arglist,pVM vm)
{
   int64_t z = zToI64(self->bi);
   BigInt *r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);

   if (z<0) z = 0;
   mpz_init(r->bi);
   mpz_fac_ui(r->bi,(unsigned long)z);
   return (Instance *)r;
}

    // .set(n): modify in place
static Instance *BI_set(BigInt *self,pArglist arglist,pVM vm)
{
   Instance *n = arglistGet(arglist,0,0,vm);

   if (IS_PtrInt(n))
   {
      mpz_init_set_si(self->bi,PtrInt_TO_N(n));
      return (Instance *)self;
   }
   if (n->objID == bigIntId)
   {
      mpz_set(self->bi,BIbi(n));
      return (Instance *)self;
   }

   switch(TYPEO1(n))
   {
      case IntType:
      {
	 i64ToZ(INT64V(n),self->bi);
	 break;
      }
      default: vmThrow(vm,E_TYPE_ERROR,"BigNum.set: No");
   }

   return (Instance *)self;
}

// .add(), .sub(), .mul(), .div(), .mod()
static Instance *BI_iadd(BigInt *self,pArglist arglist,pVM vm);
static Instance *BI_isub(BigInt *self,pArglist arglist,pVM vm);
static Instance *BI_imul(BigInt *self,pArglist arglist,pVM vm);
static Instance *BI_idiv(BigInt *self,pArglist arglist,pVM vm);
static Instance *BI_imod(BigInt *self,pArglist arglist,pVM vm);
static Instance *BI_divisibleBy(BigInt *self,pArglist arglist,pVM vm)
{
   int      r;
   int64_t  N = arglistGetInt(arglist,0,0,vm);
   BigInt  *pN=0;

   if (pN) r = mpz_divisible_p(self->bi,pN->bi);
   else    r = mpz_divisible_ui_p(self->bi,(unsigned long)N);
   return r ? BoolTrue : BoolFalse;
}

    // .copy(): make a copy. Nice for initialization: pump(...,BN(1).copy)
static Instance *BI_copy(Instance *self,pArglist arglist,pVM vm)
{
   BigInt *bi = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_init_set(bi->bi,BIbi(self));
   return (Instance *)bi;
}

    // .probablyPrime(reps=25): 
static Instance *BI_probablyPrime(BigInt *self,pArglist arglist,pVM vm)
{
   int     reps=25;
   int64_t i64;

   if (arglistTryToGetInt(arglist,0,&i64,0,vm)) reps=(int)i64;
   return mpz_probab_prime_p(self->bi,reps) ? BoolTrue : BoolFalse;
}

    // BI.nextPrime(n): in place
static Instance *BI_nextPrime(BigInt *self,pArglist arglist,pVM vm)
{
   mpz_nextprime(self->bi,self->bi);
   return (Instance *)self;
}

    // BI.binomial(k): -->BN(binomial(self,k))
static Instance *BI_binomial(Instance *self,pArglist arglist,pVM vm)
{
   int64_t k  = arglistGetInt(arglist,0,0,vm);
   BigInt *bi = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_init(bi->bi);
   mpz_bin_ui(bi->bi,BIbi(self),(unsigned long int)k);
   return (Instance *)bi;
}

static int BI_cmp2(BigInt *self,Instance *X,pVM);

    // BI.min(n)
static Instance *BI_min(Instance *self,pArglist arglist,pVM vm)
{
   Instance *X = arglistGet(arglist,0,0,vm);
   if(BI_cmp2((BigInt *)self,X,vm)<=0) return self;
   return X;
}
    // BI.max(n)
static Instance *BI_max(Instance *self,pArglist arglist,pVM vm)
{
   Instance *X = arglistGet(arglist,0,0,vm);
   if(BI_cmp2((BigInt *)self,X,vm)>=0) return self;
   return X;
}

    // bi.isPerfectSquare() --> Bool
    //!!! add isPerfectPower(list)
static Instance *BI_isPerfectSquare(BigInt *self,pArglist arglist,pVM vm)
   { return mpz_perfect_power_p(self->bi) ? BoolTrue : BoolFalse; } 

    // bi.isPerfectPower() --> Bool
    //!!! add isPerfectPower(list)
static Instance *BI_isPerfectPower(BigInt *self,pArglist arglist,pVM vm)
   { return mpz_perfect_power_p(self->bi) ? BoolTrue : BoolFalse; } 

    // bi.jacobi(p) --> Jacobi symbol (self | p) --> Void if p not odd
static Instance *BI_jacobi(BigInt *self,pArglist arglist,pVM vm)
{ 
   Instance *ip = arglistGet(arglist,0,0,vm);
   int64_t   p;

   if (IS_PtrInt(ip))
   {
      mpz_t mp;
      p = PtrInt_TO_N(ip);
    p:
      if(p%2==0) return Void;	// only defined for p odd.
      mpz_init_set_ui(mp,p);
      return intCreate(mpz_jacobi(self->bi,mp),vm);
   }
   if (TYPEO1(ip)== IntType) { p = INT64V(ip); goto p; }
   if (ip->objID == bigIntId)
   {
      if(mpz_even_p(((BigInt *)ip)->bi)) return Void;
      return intCreate(mpz_jacobi(self->bi,((BigInt *)ip)->bi),vm);
   }
   return Void;
}


static MethodTable biMethods[] =
{
   "create",		(pMethod)BN_create,
   "toInt",		(pMethod)BI_toInt,
   "toFloat",		(pMethod)BI_toFloat,
   "toString",		(pMethod)BI_toString,
   "toIntString",	(pMethod)BI_toIntString,
   "toData",		(pMethod)BI_toData,
   "unbox",		(pMethod)BI_unbox,
   "len",		(pMethod)BI_len,
   "toBool",		(pMethod)BI_toBool,
   "abs",		(pMethod)BI_abs,
   "bitAnd",		(pMethod)BI_bitAnd,
   "bitOr",		(pMethod)BI_bitOr,
   "bitXor",		(pMethod)BI_bitXor,
   "bitNot",		(pMethod)BI_bitNot,
   "shiftLeft",		(pMethod)BI_shiftLeft,
   "shiftRight",	(pMethod)BI_shiftRight,
   "gcd",		(pMethod)BI_gcd,
   "gcdExt",		(pMethod)BI_gcdExt,
   "lcm",		(pMethod)BI_lcm,
   "div2",		(pMethod)BI_div2,
   "divr",		(pMethod)BI_div2,
   "pow",		(pMethod)BI_pow,
   "powm",		(pMethod)BI_powm,
   "root",		(pMethod)BI_root,
   "sqrt",		(pMethod)BI_sqrt,
   "factorial",		(pMethod)BI_factorial,
   "probablyPrime",	(pMethod)BI_probablyPrime,
   "nextPrime",		(pMethod)BI_nextPrime,
   "binomial",		(pMethod)BI_binomial,
   "min",		BI_min,
   "max",		BI_max,
   "isPerfectSquare",	(pMethod)BI_isPerfectSquare,
   "isPerfectPower",	(pMethod)BI_isPerfectPower,
   "jacobi",		(pMethod)BI_jacobi,

   "set",		(pMethod)BI_set,
   "add",		(pMethod)BI_iadd,
   "sub",		(pMethod)BI_isub,
   "mul",		(pMethod)BI_imul,
   "div",		(pMethod)BI_idiv,
   "mod",		(pMethod)BI_imod,
   "divisibleBy",	(pMethod)BI_divisibleBy,

   "copy",		BI_copy,

   0,			0
};


	/////////////////////////// Properties
    // .isEven: is self even?
static Instance *BI_isEven(Instance *self, pVM vm)
   { return mpz_even_p(((BigInt *)self)->bi) ? BoolTrue : BoolFalse; }

    // .isOdd
static Instance *BI_isOdd(Instance *self, pVM vm)
   { return mpz_odd_p(((BigInt *)self)->bi) ? BoolTrue : BoolFalse; }

    // .numDigits --> num digits in self. Zero has no digits
static Instance *BI_numDigits(Instance *self, pVM vm)
   { return intCreate(biNumDigits(((BigInt *)self)->bi,10,vm),vm); }

    // bi.num1s --> number of 1 bits in self
static Instance *BI_num1s(BigInt *self,pVM vm)
   { return intCreate(mpz_popcount(self->bi),vm); }

    // .sign: >0 --> 1, 0 --> 0, <0 --> -1
static Instance *BI_sign(Instance *self, pVM vm)
{
   int s = mpz_cmp_si(((BigInt *)self)->bi,0);
   if (s < 0) return MinusOne;
   if (s > 0) return One;
   return Zero;
}

static const PropertyTable biProperties[] =
{
   "sign",	(pProperty)BI_sign,
   "isEven",	(pProperty)BI_isEven,
   "isOdd",	(pProperty)BI_isOdd,
   "numDigits",	(pProperty)BI_numDigits,
   "num1s",	(pProperty)BI_num1s,
   0,		0
};



	/////////////////////////// ops
static Instance *biAdd(BigInt *self,Instance *X,int alloc,pVM vm)
{
   int64_t  N;
   BigInt  *pN, *r = eyeballIX(X,&N,&pN,alloc,vm);
   if (!r) r = self;

   if (pN) mpz_add(r->bi,self->bi,pN->bi);
   else
   {
      mpz_t z;
      i64ToZ(N,z);
      mpz_add(r->bi,self->bi,z); mpz_clear(z);
   }
   return (Instance *)r;
}
static Instance *BI_add(BigInt *self,Instance *X,pVM vm)
   { return biAdd(self,X,1,vm); }
static Instance *BI_iadd(BigInt *self,pArglist arglist,pVM vm)
   { return biAdd(self,arglistGet(arglist,0,0,vm),0,vm); }

static Instance *biSub(BigInt *self,Instance *X,int alloc,pVM vm)
{
   int64_t  N;
   BigInt  *pN, *r = eyeballIX(X,&N,&pN,alloc,vm);
   if (!r) r = self;

   if (pN) mpz_sub(r->bi,self->bi,pN->bi);
   else
   {
      mpz_t z;
      i64ToZ(N,z);
      mpz_sub(r->bi,self->bi,z); mpz_clear(z);
   }
   return (Instance *)r;
}
static Instance *BI_sub(BigInt *self,Instance *X,pVM vm)
   { return biSub(self,X,1,vm); }
static Instance *BI_isub(BigInt *self,pArglist arglist,pVM vm)
   { return biSub(self,arglistGet(arglist,0,0,vm),0,vm); }

static Instance *biMul(BigInt *self,Instance *X,int alloc,pVM vm)
{
   int64_t  N;
   BigInt  *pN, *r = eyeballIX(X,&N,&pN,alloc,vm);
   if (!r) r = self;

   if (pN) mpz_mul(r->bi,self->bi,pN->bi);
   else
   {
      #if (LONG_MAX == 2147483647L)	// 32 bit ints
	 mpz_t z;
	 i64ToZ(N,z);
	 mpz_mul(r->bi,self->bi,z); mpz_clear(z);
      #else
	 mpz_mul_si(r->bi,self->bi,N);
      #endif
   }
   return (Instance *)r;
}
static Instance *BI_mul(BigInt *self,Instance *X,pVM vm)
   { return biMul(self,X,1,vm); }
static Instance *BI_imul(BigInt *self,pArglist arglist,pVM vm)
   { return biMul(self,arglistGet(arglist,0,0,vm),0,vm); }

static Instance *biDiv(BigInt *self,Instance *X,int alloc,pVM vm)
{
   int64_t  N;
   BigInt  *pN, *r = eyeballIX(X,&N,&pN,alloc,vm);
   if (!r) r = self;

   if (pN)
   {
      if (0 == mpz_cmp_ui(pN->bi,0)) vmThrow(vm,E_MATH_ERROR,"Divide by zero");
      mpz_tdiv_q(r->bi,self->bi,pN->bi);
   }
   else
   {
      mpz_t d;
      if (N == 0) vmThrow(vm,E_MATH_ERROR,"Divide by zero");
      i64ToZ(N,d);
      mpz_tdiv_q(r->bi, self->bi,d); mpz_clear(d);
   }
   return (Instance *)r;
}
static Instance *BI_div(BigInt *self,Instance *X,pVM vm)
   { return biDiv(self,X,1,vm); }
static Instance *BI_idiv(BigInt *self,pArglist arglist,pVM vm)
   { return biDiv(self,arglistGet(arglist,0,0,vm),0,vm); }

static Instance *biMod(BigInt *self,Instance *X,int alloc,pVM vm)
{
   int64_t  N;
   BigInt  *pN, *r = eyeballIX(X,&N,&pN,alloc,vm);
   if (!r) r = self;

   if (pN)
   {
      if (0 == mpz_cmp_ui(pN->bi,0)) mpz_set_ui(r->bi,0);
      else			     mpz_mod(r->bi,self->bi,pN->bi);
   }
   else
   {
      mpz_t d;
      if (N == 0) mpz_set_ui(r->bi,0);
      else
      {
	 i64ToZ(N,d);
	 mpz_mod(r->bi, self->bi,d); mpz_clear(d);
      }
   }
   return (Instance *)r;
}
static Instance *BI_mod(BigInt *self,Instance *X,pVM vm)
   { return biMod(self,X,1,vm); }
static Instance *BI_imod(BigInt *self,pArglist arglist,pVM vm)
   { return biMod(self,arglistGet(arglist,0,0,vm),0,vm); }

static Instance *BI_negate(BigInt *self,Instance *X,pVM vm)
{
   BigInt *r = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);

   mpz_init(r->bi);
   mpz_neg(r->bi,self->bi);
//   return addToCollectables((Instance *)r,I_OWNED,vm);
   return (Instance *)r;
}

static int BI_cmp(BigInt *self,Instance *X,pVM vm)
{
   if (IS_PtrInt(X)) return mpz_cmp_si(self->bi,PtrInt_TO_N(X));
   if (TYPEO1(X) == IntType)
   {
      #if (LONG_MAX == 2147483647L)	// 32 bit ints
         int   s;
         mpz_t z;
	 i64ToZ(INT64V(X),z);
	 s = mpz_cmp(self->bi,z); mpz_clear(z);
	 return s;
      #else
         return mpz_cmp_si(self->bi,INT64V(X));
      #endif
   }
   if (X->objID != bigIntId) return 1;	// !eq, not comparable
   return mpz_cmp(self->bi,((BigInt *)X)->bi);
}

static Instance *BI_eq(BigInt *self,Instance *X,pVM vm)
   { return (BI_cmp(self,X,vm) == 0) ? BoolTrue : BoolFalse; }
static Instance *BI_neq(BigInt *self,Instance *X,pVM vm)
   { return BI_cmp(self,X,vm) ? BoolTrue : BoolFalse; }

static int BI_cmp2(BigInt *self,Instance *X,pVM vm)
{
   if (IS_PtrInt(X)) return mpz_cmp_si(self->bi,PtrInt_TO_N(X));
   if (TYPEO1(X) == IntType)
   {
      #if (LONG_MAX == 2147483647L)	// 32 bit ints
         int   s;
         mpz_t z;
	 i64ToZ(INT64V(X),z);
	 s = mpz_cmp(self->bi,z); mpz_clear(z);
	 return s;
      #else
         return mpz_cmp_si(self->bi,INT64V(X));
      #endif
   }
   if (X->objID != bigIntId) // not comparable
      vmThrow(vm,E_TYPE_ERROR,"BigInt can't compare");
   return mpz_cmp(self->bi,((BigInt *)X)->bi);
}

static Instance *BI_lt(BigInt *self,Instance *X,pVM vm)
   { return (BI_cmp2(self,X,vm) < 0) ? BoolTrue : BoolFalse; }
static Instance *BI_lte(BigInt *self,Instance *X,pVM vm)
   { return (BI_cmp2(self,X,vm) <= 0) ? BoolTrue : BoolFalse; }

static Instance *BI_gt(BigInt *self,Instance *X,pVM vm)
   { return (BI_cmp2(self,X,vm) > 0) ? BoolTrue : BoolFalse; }
static Instance *BI_gte(BigInt *self,Instance *X,pVM vm)
   { return (BI_cmp2(self,X,vm) >= 0) ? BoolTrue : BoolFalse; }


static const OpcodeTable biOps[] = 
{
   OP_EQ,	(pOp)BI_eq,
   OP_NEQ,	(pOp)BI_neq,

   OP_LT,	(pOp)BI_lt,
   OP_LTE,	(pOp)BI_lte,
   OP_GT,	(pOp)BI_gt,
   OP_GTE,	(pOp)BI_gte,

   OP_ADD,	(pOp)BI_add,
   OP_SUB,	(pOp)BI_sub,
   OP_MUL,	(pOp)BI_mul,
   OP_DIV,	(pOp)BI_div,
   OP_MOD,	(pOp)BI_mod,

   OP_NEGATE,	(pOp)BI_negate,

   0,		0
};


#if 0
/* ******************************************************************** */
/* *************************** Big Float ****************************** */
/* ******************************************************************** */

//!!!!ACK! split this off & use MPFR

static int bfFree(BigFloat *self) { mpf_clear(self->bf); return 1; }


    // BigFloat.toString() --> String
    // Float.toString([precision[,format]]
static Instance *BF_toString(BigFloat *self,pArglist arglist,pVM vm)
{
#if 0
   int	     p = 12;
   int64_t   i64;
   Instance *r;
   mp_exp_t  expPtr;

   if (arglistTryToGetInt(arglist,0,&i64,0,vm)) p = (int)i64;

   r = stringAllocate(301,&ptr,vm);
   mpz_get_str(ptr,&expPtr,10,300,self->bf);
   return addToCollectables((Instance *)r,I_OWNED,vm);
#else
   char	   format[20], buf[310];
   int     precision = 6;
   int64_t i64;

   if (arglistTryToGetInt(arglist,0,&i64,0,vm)) precision = (int)i64;
   sprintf(format,"%%.%dF%c",precision,'f');
   gmp_sprintf(buf,format,self->bf);
   return stringCreate(buf,I_OWNED,vm);
#endif
}

static Instance *BF_toInt(BigFloat *self,pArglist arglist,pVM vm)
   { return intCreate(mpf_get_si(self->bf),vm); }

static Instance *BF_toFloat(BigFloat *self,pArglist arglist,pVM vm)
   { return floatCreate(mpf_get_d(self->bf),vm); }

static Instance *BF_toBool(BigFloat *self,pArglist arglist,pVM vm)
   { return mpf_cmp_ui(self->bf,0) ?  BoolTrue : BoolFalse; }

static Instance *BF_ceil(BigFloat *self,pArglist arglist,pVM vm)
   { mpf_ceil(self->bf,self->bf); return (Instance *)self; }

static Instance *BF_floor(BigFloat *self,pArglist arglist,pVM vm)
   { mpf_floor(self->bf,self->bf); return (Instance *)self; }

static Instance *BF_trunc(BigFloat *self,pArglist arglist,pVM vm)
   { mpf_trunc(self->bf,self->bf); return (Instance *)self; }

    // bf.pow(n) --> self ^ n, in place
static Instance *BF_pow(BigFloat *self,pArglist arglist,pVM vm)
{
   unsigned long N = (unsigned long)arglistGetInt(arglist,0,0,vm);

   mpf_pow_ui(self->bf,self->bf,N); 
   return (Instance *)self;
}

static Instance *BF_iadd(BigFloat *self,pArglist arglist,pVM vm);
static Instance *BF_isub(BigFloat *self,pArglist arglist,pVM vm);
static Instance *BF_imul(BigFloat *self,pArglist arglist,pVM vm);
static Instance *BF_idiv(BigFloat *self,pArglist arglist,pVM vm);
static Instance *BF_imod(BigFloat *self,pArglist arglist,pVM vm);

static MethodTable bfMethods[] =
{
   "toInt",		(pMethod)BF_toInt,
   "toFloat",		(pMethod)BF_toFloat,
   "toString",		(pMethod)BF_toString,
   "toBool",		(pMethod)BF_toBool,

   "ceil",		(pMethod)BF_ceil,
   "floor",		(pMethod)BF_floor,
   "trunc",		(pMethod)BF_trunc,

   "pow",		(pMethod)BF_pow,

//   "add",		(pMethod)BF_iadd,
//   "sub",		(pMethod)BF_isub,
   "mul",		(pMethod)BF_imul,
   "div",		(pMethod)BF_idiv,

   0,			0
};


  /////////////////////////////////////////////////////////////////////////
 ///////////////////////////// Ops //////////////////////////////////////
///////////////////////////////////////////////////////////////////////

static Instance *bfMul(BigFloat *self,Instance *X,int alloc,pVM vm)
{
   BigFloat *r = self, *x;

   if (IS_PtrInt(X) || X->objID != bigFloatId)
   	 vmThrow(vm,E_VALUE_ERROR,"BigFloat *: BigFloat only");

   x = ((BigFloat *)X);
   mpf_mul(r->bf,r->bf,x->bf);
   return (Instance *)r;
}
//static Instance *BF_mul(BigFloat *self,Instance *X,pVM vm)
//   { return bfMul(self,X,1,vm); }
static Instance *BF_imul(BigFloat *self,pArglist arglist,pVM vm)
   { return bfMul(self,arglistGet(arglist,0,0,vm),0,vm); }

static Instance *bfDiv(BigFloat *self,Instance *X,int alloc,pVM vm)
{
   BigFloat *r = self, *x;

   if (IS_PtrInt(X) || X->objID != bigFloatId)
   	 vmThrow(vm,E_VALUE_ERROR,"BigFloat /: BigFloat only");

   x = ((BigFloat *)X);
   if (0 == mpf_cmp_ui(x->bf,0)) vmThrow(vm,E_MATH_ERROR,"Divide by zero");
   mpf_div(r->bf,r->bf,x->bf);
   return (Instance *)r;
}
//static Instance *BF_div(BigFloat *self,Instance *X,pVM vm)
//   { return bfDiv(self,X,1,vm); }
static Instance *BF_idiv(BigFloat *self,pArglist arglist,pVM vm)
   { return bfDiv(self,arglistGet(arglist,0,0,vm),0,vm); }



static const OpcodeTable bfOps[] = 
{
#if 0
   OP_EQ,	(pOp)BF_eq,
   OP_NEQ,	(pOp)BF_neq,

   OP_LT,	(pOp)BF_lt,
   OP_LTE,	(pOp)BF_lte,
   OP_GT,	(pOp)BF_gt,
   OP_GTE,	(pOp)BF_gte,

   OP_ADD,	(pOp)BF_add,
   OP_SUB,	(pOp)BF_sub,
   OP_MUL,	(pOp)BF_mul,
   OP_DIV,	(pOp)BF_div,
   OP_MOD,	(pOp)BF_mod,

   OP_NEGATE,	(pOp)BF_negate,
#endif
   0,		0
};

#endif

/* ******************************************************************** */
/* ****************************** Big Num ***************************** */
/* ******************************************************************** */

static ZKL_Object BigNumObject;

static gmp_randstate_t randomState;
static SpinLock	       rlock;

#if (LONG_MAX == 2147483647L)	// 32 bit ints
   #define RMAX	  ULONG_MAX
   #define RFMAX  ULONG_MAX
#else				// assume 64 bits ints (or bigger)
   #define RMAX  9223372036854775807LL	// LONG_MAX
      // A double has 52 bits for the mantissa, give myself some room when
      // moving from ints to doubles 
   #define RFMAX 1125899906842624LL	// 2^50, fits in double
#endif

    // random() --> random non negative integer
    /* random([start, stop])
     * Return a random number, optionally in [start,stop)
     *    start == int   --> int
     *    start == float --> float
     */
static Instance *randomNum(Instance *self,pArglist arglist,pVM vm)
{
   Instance *x;
   unsigned long int R;

   if ((x = arglistTryToGet(arglist,0)))
   {
      if (IS_PtrInt(x) || TYPEO1(x) == IntType)
      {
	 int64_t a = convertToInt(x,vm);
	 int64_t b = arglistGetInt(arglist,1,0,vm);
	 mpz_t   z;

	 i64ToZ(b - a,z);
	 SPIN_LOCK_ACQUIRE2(&rlock);
	    mpz_urandomm(z,randomState,z);	 // 0 <= z < z
	 SPIN_LOCK_RELEASE(&rlock);
	 x = intCreate(a + zToI64(z),vm); mpz_clear(z);
	 return x;
      }
      else	// float
      {
	 double	a,b,width, r;
	 a = arglistGetFloat(arglist,0,0,vm);
	 b = arglistGetFloat(arglist,1,0,vm);
	 SPIN_LOCK_ACQUIRE2(&rlock);
	    R = gmp_urandomm_ui(randomState,RFMAX);
	 SPIN_LOCK_RELEASE(&rlock);
	 if (b < a) { r = b; b = a; a = r; }
	 width = b - a;
	 r = a + ((double)R / ((double)RFMAX + 1.0) * width);
	 return floatCreate(r,vm);
      }
   }

   SPIN_LOCK_ACQUIRE2(&rlock);
      R = gmp_urandomm_ui(randomState,RMAX);
   SPIN_LOCK_RELEASE(&rlock);
   return intCreate(R,vm);
}

    // setRandomSeed(newSeed) --> Void
static Instance *setRandomSeed(Instance *self,pArglist arglist,pVM vm)
{
   unsigned long newSeed = (unsigned long)arglistGetInt(arglist,0,0,vm);
   SPIN_LOCK_ACQUIRE2(&rlock);
      gmp_randseed_ui(randomState,newSeed);
   SPIN_LOCK_RELEASE(&rlock);
   return Void;
}

    // .create(123)   --> BigInt
    // .create("123" [,base])  --> BigInt
    // .create(Data) --> bytes to BigInt
    // .create(BigInt N) --> N
static Instance *BN_create(Instance *self,pArglist arglist,pVM vm)
{
   Instance *n = arglistGet(arglist,0,0,vm);
   int       s;

   if (IS_PtrInt(n))
   {
      BigInt *bi = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
      mpz_init_set_si(bi->bi,PtrInt_TO_N(n));
      return (Instance *)bi;
   }

   s = n->objID;
//   if (s==bigIntId || s==bigFloatId) return n;
   if (s==bigIntId) return BI_copy(n,NoArglist,vm);  // BN(BN)-->copy

   switch(TYPEO1(n))
   {
      case StringType:
      {
	 int       base = 10;
	 Instance *i;
	 char     *ns = stringText(n), *ptr;
	 mpz_t	   *bi;

	 if ((i = arglistTryToGet(arglist,1)))	// base or Void?
	 {
#if 0
	    if (i == Void)	// Void --> Float
	    {
	       i = instanceAllocate(sizeof(BigFloat),&BigFloatObject,1,vm);
addToCollectables(i,I_OWNED,vm);	// gc on error
	       s = mpf_init_set_str(((BigFloat *)i)->bf,stringText(n),10);
	       goto finishIt;
	    }
	    else 
#endif
	    base = (int)convertToInt(i,vm);	// and fall through
	 }
	 i  = ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
	 bi = &((BigInt *)i)->bi;
	 
	 if(base==10 && (ptr = strchr(ns,'e')))	// "10e100"
	 {	// what is 0x123e10? 1195536
	    mpz_t be;

	    *ptr = '\0';
   	    s = mpz_init_set_str(*bi,ns,base);
	    *ptr = 'e';  // restore instance
	    if (s == -1) vmThrow(vm,E_VALUE_ERROR,"BigNum(string): Bad character");
	    mpz_init_set_ui(be,10); mpz_pow_ui(be, be, atoi(ptr+1));
	    mpz_mul(*bi,*bi,be);
	    mpz_clear(be);	// GC
	 }
	 else s = mpz_init_set_str(*bi,ns,base);
//      finishIt:
//	 addToCollectables(i,I_OWNED,vm);	// gc on error
	 if (s == -1) vmThrow(vm,E_VALUE_ERROR,"BigNum(string): Bad character");
	 return i;
      }
      case IntType:
      {
	 BigInt *i = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);

	 i64ToZ(INT64V(n),i->bi);
	 n = (Instance *)i;
	 break;
      }
      case DataType:
      {
	 BigInt *i = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
	 size_t  sz;
	 int	 endian = 0, order = 1;  // order is endian-ness of the wad
	 Byte   *bytes = dataText(n,&sz);

	 mpz_import(i->bi,sz,order,1,endian,0,bytes);
	 n = (Instance *)i;
	 break;
      }
#if 0
      case FloatType:
      {
	 BigFloat *f = (BigFloat *)
		 instanceAllocate(sizeof(BigFloat),&BigFloatObject,1,vm);
	 mpf_init_set_d(f->bf,FLOATV(n)); 
	 n = (Instance *)f;
addToCollectables(n,I_OWNED,vm);
	 break;
      }
#endif
      default: vmThrow(vm,E_TYPE_ERROR,"BigNum: No");
   }

   return n;
}

    // BN.primorial(n): -->mul(primes <= n)
static Instance *BN_primorial(Instance *self,pArglist arglist,pVM vm)
{
   int64_t n  = arglistGetInt(arglist,0,0,vm);
   BigInt *bi = (BigInt *)ibucketAllocate(biBuckets,&BigIntObject,I_OWNED,1,vm);
   mpz_init(bi->bi);
   mpz_primorial_ui(bi->bi,(unsigned long int)n);
   return (Instance *)bi;
}


static MethodTable methods[] =
{
   "create",		(pMethod)BN_create,
   "random",		(pMethod)randomNum,
   "setRandomSeed",	(pMethod)setRandomSeed,
   "primorial",		BN_primorial,
   "toBool",		Bool_soTrue,
   0,			0
};



#if 0
    // .version
static Instance *version(Instance *self, pVM vm)
   { return intCreate(lzo_version(),vm); }	// 0x2020

    // .versionString
static Instance *versionString(Instance *self, pVM vm)
   { return stringCreate((char *)lzo_version_string(),I_OWNED,vm); }
#endif

    // .authors
static Instance *authors(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,"Craig Durland","http://zenkinetic.com/",
          (char *)0); 
}

    // *.libraries
static Instance *libraries(Instance *self, pVM vm)
{
   return createTupleOfStrings(vm,
      "MPIR","http://www.mpir.org/",
      "GMP", "http://gmplib.org/",
      (char *)0);
}

static PropertyTable properties[] =
{
//   "version",		(pProperty)version,
//   "versionString",	(pProperty)versionString,
//   "versionDate",	(pProperty)versionDate,
   "authors",		(pProperty)authors,
   "libraries",		(pProperty)libraries,
   0,			0
};



Instance bigNumInstance;

//static pMethod in_biMethods_methods(Instance *ignore, register char *str);

DllExport void *construct(void *vm)
{
   static IBucketHeader _biBuckets;	// might not be used

   if (!vm) return (void *)ZKLX_PROTOCOL;	// handshake
   	// If this is reloaded, nothing happens except 
	// construct() is called again so don't reinitialize
   if (!BigNumObject.name)	// static items are zero
   {
      constructObject(&BigNumObject,NativeType, methods,properties,0,vm);
      BigNumObject.name	      = "BigNum";
      BigNumObject.vaultPath  = "Math";
      BigNumObject.isize      = sizeof(Instance);
      BigNumObject.threadSafe = 1;

      instanceInit(&bigNumInstance,&BigNumObject,I_UNTOUCHABLE);


      constructObject(&BigIntObject,NativeType, biMethods,biProperties,biOps,vm);
      BigIntObject.name	        = "BigInt";
      BigIntObject.vaultPath    = "Math";
      BigIntObject.freeMe       = biFree;
      BigIntObject.isBInstance  = 1;
      BigIntObject.isize        = sizeof(BigInt);
      BigIntObject.threadSafe   = 1;
      bigIntId		        = BigIntObject.id;

	// share w/Int/Dictionary on 32 bit machines, Dictionary on 64
      biBuckets = ibucketHitchHike(&BigIntObject,0,4321,&_biBuckets,(pVM)vm);


#if 0
      constructObject(&BigFloatObject,NativeType, bfMethods,properties,bfOps,vm);
      BigFloatObject.name       = "BigFloat";
      BigFloatObject.freeMe     = bfFree;
      BigFloatObject.isize      = sizeof(BigFloat);
      BigFloatObject.threadSafe = 1;
      bigFloatId	        = BigFloatObject.id;

//mpf_set_default_prec(400);
#endif

      spinLockInit(&rlock);
      gmp_randinit_default(randomState);
      	// !!!! not good enough seed ?MAC
      gmp_randseed_ui(randomState,(unsigned long)time(0) + clock());
   }
   return &bigNumInstance;
}



///////////////////////////////////////////////////////////////////
// zkl extractTable.zkl < zklBigNum.c | gperf | zkl gperf.zkl -i biMethods

